     { Copyright (C) 1984 D.M. Fritz-Rohner                }

program main ;

     { Name:                                               }
     {     FLUFF - Error FLUFFing Algorithm.               }
     {                                                     }
     { Version:                       Date:                }
     {     1.0/TURBO (tm)                  1984 July 7     }
     {                                                     }
     { Purpose:                                            }
     {     Demonstrate PASCAL application of error fluff-  }
     {     ing algorithm proposed by the reference.        }
     {                                                     }
     { Author:                                             }
     {     D.M. Fritz-Rohner                               }
     {     Post Office Box 9080                            }
     {     Akron, Ohio 44305                               }
     {                                                     }
     { Reference:                                          }
     {     'A Simple Minimax Algorithm'                    }
     {     Steven A. Ruzinsky                              }
     {     Dr. Dobbs Journal                               }
     {     Number 93, July, 1984                           }
     {     pg. 84 et seq                                   }
     {                                                     }
     { Description:                                        }
     {     This program estimates polynomial coefficients  }
     {     for a minimax fit to an elementary transcenden- }
     {     tal function.  See FLUFF10.DOC.                 }
     {                                                     }
     {     This version is written in TURBO Pascal and is  }
     {     transliterated from the Basic version pre-      }
     {     sented in the reference.  See FLUFF10.BAS.      }
     {                                                     }
     {     The variable t is substituted for intermediate  }
     {     variable usage of the symbols qx and xk.  Ex-   }
     {     ponentiating operator usage is circumvented.    }
     {     Covariance matrix inverse and coefficient       }
     {     vector are fully initialized.                   }
     {                                                     }
     { TURBO Pascal (tm) Borland International             }

const
     N              = 3 ;
     M              = 50 ;
     Iterations     = 1000 ;
var
     x              : array[1..M,1..N] of real ;
     y              : array[1..M] of real ;
     xk             : array[1..N] of real ;
     q              : array[1..N,1..N] of real ;
     qx             : array[1..N] of real ;
     ak             : array[1..N] of real ;
     a              : array[1..N] of real ;

     d              : real ;
     e, e2          : real ;
     eMax           : real ;
     eBest          : real ;
     t              : real ;
     i, j, k        : integer ;
     jMax           : integer ;

{~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~}

procedure IncludeData ;
     { Incorporate each data point once.                   }
begin
   e := y[k] ;
   for i := 1 to N do begin
      t := x[k,i] ;
      xk[i] := t ;
      e := e - ak[i]*t
   end
end ;

{~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~}

procedure SaveBest ;
     { Save coefficients of best fit.                      }
begin
   for i := 1 to N do
      a[i] := ak[i]
end ;

{~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~}

procedure FindMax ;
     { Find maximum absolute error and corresponding index. }
begin
   eMax := 0.0 ;
   jMax := 1 ;
   for j := 1 to M do begin
      e := y[j] ;
      for i := 1 to N do
         e := e - x[j,i]*ak[i] ;
      e := ABS(e) ;
      if e > eMax then begin
         eMax := e ;
         jMax := j
      end
   end ;
   writeln('Iterations = ',k-M:4,'   Maximum Absolute Error = ',eMax:12) ;
   e := y[jMax] ;
   for i := 1 to N do begin
      t := x[jMax,i] ;
      xk[i] := t ;
      e := e - ak[i]*t
   end ;
   if eMax < eBest then begin
      eBest := eMax ;
      SaveBest
   end
end ;

{~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~}

begin

     { Fill y and x with data. The function used           }
     { is SIN(Pi*x/2), however, the data is modified so    }
     { that the minimax criterion is applied to the rela-  }
     { tive error.                                         }

   for i := 1 to M do begin
      y[i] := 1.0 ;
      e := i/M ;
      e2 := e*e ;
      d := 1.0/SIN(1.570796327*e) ;
      for j := 1 to N do begin
         x[i,j] := d*e ;
         e := e*e2
      end
   end ;

     { The following initiates the q matrix.  It may be    }
     { necessary to to adjust the number below for         }
     { best results.                                       }

   for i := 1 to N do begin
      q[i,i] := 1.0e+06 ;
      if i < N then begin
         for j := i+1 to N do begin
            q[i,j] := 0.0 ;
            q[j,i] := 0.0
         end
      end ;
      xk[i] := 0.0 ;
      ak[i] := 0.0
   end ;

     { The following loop with index, k, reiterates the    }
     { sequential least squares algorithm. Up to limit M,  }
     { each data point is incorporated once into q and ak. }
     { This results in a least squares fit to the data.    }
     { Afterwards, the data corresponding to the  maximum  }
     { absolute error are incorporated back into q and ak. }

   eBest := 1.0 ;
   for k := 1 to Iterations+M do begin
      if k > M then
         FindMax
      else
         IncludeData ;
      d := 1.0 ;
      for j := 1 to N do begin
         t := 0.0 ;
         for i := 1 to N do
            t := t + xk[i]*q[j,i] ;
         qx[j] := t ;
         d := d + xk[j]*t ;
      end ;
      for j := 1 to N do begin
         t := qx[j]/d ;
         for i := 1 to N do
            q[i,j] := q[i,j] - qx[i]*t ;
      end ;
      for j := 1 to N do begin
         t := 0.0 ;
         for i := 1 to N do
            t := t + xk[i]*q[j,i] ;
         ak[j] := ak[j] + t*e
      end
   end ;

     { Print the results.                                  }

   writeln('Coefficients:') ;
   for i := 1 to N do begin
      ak[i] := a[i] ;
      writeln(ak[i]:12)
   end ;
   FindMax 
end.
